<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">

	<!-- new client & reseller's contact name prefixes -->
	<xsl:param name='client-name-prefix'>client </xsl:param>
	<xsl:param name='reseller-name-prefix'>reseller </xsl:param>
	<!-- maximum significant value for the disk space / traffic limit.
	values exceeding that would be counted as 'unlimited' -->
    <xsl:variable name='max_limit_value' select='format-number(999999*1024*1024, "#")'/>
    <xsl:variable name='agent_name' select='cPanel9dump/@agent-name'/>

	<xsl:output indent="yes" encoding="UTF-8"/>

	<xsl:template name="cPanel9dump" match="cPanel9dump">
      <migration-dump agent-name="{$agent_name}">
			<!-- convert cPanel resellers into the Plesk clients with
			domain-creation capability -->
			<xsl:apply-templates select="reseller[@name!='']"/>
			<!-- convert cPanel accounts, belonging to default reseller to Plesk
			clients without domain creation capability -->
			<xsl:apply-templates select="reseller[@name='']/account"/>
		</migration-dump>
	</xsl:template>

	<xsl:template match="ip">
		<xsl:choose>
			<xsl:when test="string-length(.) = 0">
				<!--
				Sometimes value of IP address can't be determined (disabled domains?)
				Assign arbitrary shared ip address in this case.
				-->
				<xsl:apply-templates select="../../account[
						domain/anonftp/@pub != 'true' or not(domain/anonftp/@pub)
						]/ip[string-length(.) != 0]"/>
			</xsl:when>
			<xsl:otherwise>
				<ip>
					<xsl:choose>
						<xsl:when test="../domain/anonftp/@pub='true'">
							<ip-type>exclusive</ip-type>
						</xsl:when>
						<xsl:otherwise>
							<ip-type>shared</ip-type>
						</xsl:otherwise>
					</xsl:choose>
					<ip-address>
						<xsl:value-of select='.'/>
					</ip-address>
				</ip>
			</xsl:otherwise>
		</xsl:choose>
	</xsl:template>

	<!--
	Key used for removal of duplicate IP addresses in reseller's pool:
	[IP address value][name of the upper reseller][indication of enabled anonymous ftp]
	-->
	<xsl:key name="ip" match="account/ip"
		use="concat(., ../../@name, ../domain/anonftp/@pub[.='true'])"/>

	<!--
	Convert cPanel resellers into Plesk's clients with domain-creation capabilities.
	Resold resellers accounts are converted to domain users.
	This template do not handle default reseller.
	-->
	<xsl:template name="reseller" match="reseller[string-length(@name)>0]">

		<xsl:variable name="name" select="@name"/>
		<xsl:if test='string-length($name)=0'>
			<xsl:message terminate="yes">Can&apos;t determine the reseller&apos;s name</xsl:message>
		</xsl:if>

		<client name="{$name}" contact="{concat($reseller-name-prefix, $name)}">
			<xsl:call-template name='password'>
				<xsl:with-param name='password' select='account[@name=$name]/domain/ftpuser[@name=$name]/@password'/>
			</xsl:call-template>
			<status>
				<enabled/>
			</status>
			<xsl:variable name="email" select="account[@name=$name]/@email"/>
			<xsl:if test='string-length($email)!=0'>
				<pinfo name="email">
					<xsl:value-of select="$email"/>
				</pinfo>
			</xsl:if>

			<xsl:call-template name="reseller-limits"/>
			<xsl:call-template name="reseller-permissions"/>

			<ip_pool>
				<!-- without duplicates -->
				<xsl:apply-templates select="account/ip[generate-id(.) = generate-id(key('ip',
					concat(., ../../@name, ../domain/anonftp/@pub[.='true'])))]"/>
			</ip_pool>

			<xsl:for-each select="account">
				<xsl:variable name='account-name' select='@name'/>
				<xsl:for-each select='domain'>
					<xsl:call-template name="domain">
						<xsl:with-param name="quota" select="../limit[@name='disk_space']"/>
						<xsl:with-param name="du_passwd" select="ftpuser[@name=$account-name and @name!=$name]/@password"/>
						<xsl:with-param name="du_email" select="../@email[../@name!=$name]"/>
						<xsl:with-param name="du_contact" select="../@name[.!=$name]"/>
					</xsl:call-template>
				</xsl:for-each>
			</xsl:for-each>

		</client>
	</xsl:template>

	<!--
	Converts account into client (used for default reseller's accounts)
	-->
	<xsl:template name="client" match="account">
		<xsl:variable name='name' select='@name'/>
		<client name="{$name}" contact="{concat($client-name-prefix, @name)}">

			<xsl:call-template name='password'>
				<xsl:with-param name='password' select='domain/ftpuser[@name=$name]/@password'/>
			</xsl:call-template>
			<status>
				<enabled/>
			</status>
			<xsl:if test='string-length(@email) > 0'>
				<pinfo name="email">
					<xsl:value-of select="@email"/>
				</pinfo>
			</xsl:if>

			<xsl:call-template name="account-limits"/>
			<xsl:call-template name="account-permissions"/>
			<xsl:call-template name="account-feature-permissions"/>

			<ip_pool>
				<xsl:apply-templates select="ip"/>
			</ip_pool>

			<xsl:for-each select="domain">
				<xsl:call-template name="domain">
					<xsl:with-param name="quota" select="../limit[@name='disk_space']"/>
				</xsl:call-template>
			</xsl:for-each>
		</client>
	</xsl:template>

      <!-- Handle account limits -->
      <xsl:template name="account-limits">

        <!-- Some cPanel limits may be directly mapped into Plesk's -->
        <xsl:copy-of select="limit[@name='max_traffic']"  />
        <xsl:copy-of select="limit[@name='max_maillists']"/>
        <xsl:copy-of select="limit[@name='max_box']"      />
        <xsl:copy-of select="limit[@name='max_db']"       />
        <xsl:copy-of select="limit[@name='max_subdom']"   />
        <xsl:copy-of select="limit[@name='disk_space']"   />

        <!-- Limit on ftp users is transformed to limit on webusers -->
        <xsl:if test="limit[@name='max_ftpusers']">
          <xsl:call-template name="limit">
            <xsl:with-param name="limit" select='"max_wu"'/>
            <xsl:with-param name="value" select="limit[@name='max_ftpusers']"/>
          </xsl:call-template>
        </xsl:if>
        <!--
            Limit on addon domains doesn't mean anything in Plesk
            And therefore silently skipped.
            Limit on the number of parked domains is ignored because Plesk
            domain has no limitations on number of domain aliases.
        -->

      </xsl:template>

      <!--
           Outputs domain limits setted to zero (used for addon domains).
		   Disk space is set to 50 Kb (for convenience)
		   If parameter 'skip-mail' is specified - do not
		   outputs the 'max_box' limit (used for addon domains)
      -->
      <xsl:template name='domain-limits-zero'>
		<xsl:param name='skip-mail'/>
        <limit name="max_traffic">0</limit>
        <limit name="max_maillists">0</limit>
        <xsl:if test="'' = $skip-mail">
          <limit name="max_box">0</limit>
		</xsl:if>
        <limit name="max_db">0</limit>
        <limit name="max_subdom">0</limit>
        <limit name="disk_space">51200</limit>
        <limit name="max_wu">0</limit>
      </xsl:template>

      <!--
           "Clever" domain limits processing for those limits, actual usage of what
           can't be calculated (traffic/disk space) - tries to satisfy reseller's
           limits by dividing available reseller's resources among domains, which
           have not such limit specified.

           Should be applied inside <domain>
           Default reseller has no any limits specified so this scheme works fine
           for default reseller's accounts.

           Parameters:
           name - limit name
           -->
      <xsl:template name='domain-limit-shared'>

        <xsl:param name='name'/>

        <!--
             During the domain limits processing there are following situations may
             occure:

             1 reseller has limited traffic or disk space.
             1.1 domain-owning account has its own limit (copy account's limit
             value into domain)
             1.2 domain-owning account has no limit specified or limit is unlimited
             (in this case we should divide free reseller's limit among such
             domains)

             2 reseller has no limits for traffic or disk space
             2.1 domain-owning account has its own limit (copy the limit's value)
             2.2 domain-owning account has no limit specified or limit is unlimited
             (don't doing anything - reseller is unlimited in resources anyway)

             Algorithm:
             1. if resource is limited in account - copy it's value
             2. if resource is unlimited in account - check for reseller's
             limitations and divide available reseller's resources among such
             accounts
             -->

        <xsl:variable name='limit' select='format-number(../limit[@name=$name], "#")'/>

        <xsl:choose>

          <!-- account has it's own significant limits. NOTE: as Plesk anyway
               doesn't allow limit more then 10^9 and as cPanel doesn't have
               official unlimited value (very large numbers are used for it)
               we suppose limits more than $max_limit_value as unlimited -->
          <xsl:when test='($limit != "") and ($limit != -1) and ($limit &lt; $max_limit_value)'>

            <!-- leave some space for addon domains (50K for each) -->
            <xsl:variable name='addon'  select='count(subdomain/addondomain)'/>
            <xsl:variable name='space'  select='$limit - 50*1024*$addon'/>

            <xsl:choose>
              <!-- if there is enough space for reservation for addon domains -->
              <xsl:when test='$space > 0'>
                <xsl:call-template name='limit'>
                  <xsl:with-param name='limit' select='$name'/>
                  <xsl:with-param name='value' select='format-number($space, "#")'/>
                </xsl:call-template>
              </xsl:when>
              <!-- if space isn't enough -->
              <xsl:otherwise>
                <xsl:call-template name='limit'>
                  <xsl:with-param name='limit' select='$name'/>
                  <xsl:with-param name='value' select='format-number($limit, "#")'/>
                </xsl:call-template>
              </xsl:otherwise>
            </xsl:choose>

          </xsl:when>

          <!-- either account has no limits specified, or limit is specified
               as 'unlimited' -->
          <xsl:otherwise>

            <!-- watch for reseller's limit -->
            <xsl:variable name='r_limit' select='format-number(../../limit[@name=$name], "#")'/>

            <xsl:if test='($r_limit != "") and ($r_limit != -1) and ($r_limit &lt; $max_limit_value)'>

              <!-- but reseller has the same parameter limited - so we
                   should count free reseller's resources and divide them among
                   the domains, which has no limits specified -->

              <!-- count summary reserved limits for this reseller -->
              <xsl:variable name='reserved' select='sum(../../account/limit[(@name=$name) and (. != -1) and (. != "") and (. &lt; $max_limit_value)])'/>
              <xsl:variable name='free'     select='$r_limit - $reserved'/>

              <xsl:choose>
                <!-- check that free resources are available -->
                <xsl:when test='$free &lt;= 0'>

                  <!-- reseller doesn't have enough limits to recreate
                       neccessary objects -->
                  <xsl:message>
						  <xsl:text>How do you use cPanel if it allows to exceed the reseller's "</xsl:text>
                    <xsl:value-of select='../../@name'/>
                    <xsl:text>" limit (</xsl:text>
                    <xsl:value-of select='$name'/>
                    <xsl:text> = </xsl:text>
                    <xsl:value-of select='$r_limit'/>
                    <xsl:text>) in it's resold accounts (</xsl:text>
                    <xsl:value-of select='$reserved'/>
                    <xsl:text>)? Plesk wouldn't allow it... You'll have problems with deployment of the domain </xsl:text>
                    <xsl:value-of select='@name'/>
                  </xsl:message>

                </xsl:when>

                <xsl:otherwise>
                  <!-- count a quontity of the reseller's accounts
                       (domains), having no limits specified (among which
                       reseller's limits should be divided) -->
                  <xsl:variable name='unhandled_domains'
                    select='count(../../account[(limit[@name=$name] = "") or (limit[@name=$name] = -1) or (limit[@name=$name] &gt; $max_limit_value)])'/>

                  <!-- generate limit for domain slightly less, then it
                       may	be - to have resources to create other domains
                       (standard forwarding domains) for this reseller -->

                  <!-- WARNING: trick with format-number() is needed due
                       to the bug in libxslt which causes trash output of
                       floor/round/ceiling for big numbers (10^10). When this
                       bug	eleminated it's safe to replace format-number() by
                       floor()	(Thanks to Ilnar for the sollution :) -->

                  <xsl:call-template name='limit'>
                    <xsl:with-param name='limit' select='$name'/>
                    <xsl:with-param name='value' select='format-number(($free div $unhandled_domains) - 1048576, "#")'/>
                  </xsl:call-template>

                </xsl:otherwise>

              </xsl:choose>

            </xsl:if>

          </xsl:otherwise>
        </xsl:choose>

      </xsl:template>

      <!--
           Handle reseller limits. There are 3 limits currently: number of accounts,
           traffic and disk space. Concerning traffic and disk space: there is an
           unclear situation when reseller's account has an unlimited limit, but
           reseller is limited in a number of traffic, which it can resold. What
           should I do with it? It seems preferable to limit such reseller by it's
           available reseller's limits (i.e. limit it by a number of resource it
           can resold).
           -->
      <xsl:template name="reseller-limits">
        <xsl:variable name='reseller-name' select='@name'/>

        <!-- Maximum accounts limit (if not -1) is transformed into maximum
             domains. We also should add quantity of addon domains to not
             break anything -->
        <xsl:if test="(limit[@name='max_accounts']) and (limit[@name='max_accounts'] != -1)">
          <!-- adds 1 for account of the reseller itself -->
          <xsl:variable name='account-limit' select='limit[@name="max_accounts"] + 1'/>

          <!-- add quantity of addon domains to the creation account limit -->
          <xsl:variable name="addon-domains" select="count(account/domain/subdomain/addondomain)"/>

          <xsl:call-template name="limit">
            <xsl:with-param name="limit">max_dom</xsl:with-param>
            <xsl:with-param name="value" select="$account-limit + $addon-domains"/>
          </xsl:call-template>

        </xsl:if>

        <!-- at this point we may correct the reseller's disk space and traffic
             limits -->
        <xsl:variable name='traffic_limit' select='format-number(limit[@name="max_traffic"], "#")'/>

        <xsl:if test='($traffic_limit != "") and ($traffic_limit != -1) and ($traffic_limit &lt; $max_limit_value)'>

          <!-- count summary reserved limits -->
          <xsl:variable name='reserved_traffic' select='sum(account/limit[(@name="max_traffic") and (. != -1) and (. != "") and (. &lt; $max_limit_value)])'/>

          <!-- if there is not enough resources to deploy accounts, which
               have their space reserved we should increase reseller's limit to at
               least the required value + 1Gb just in case... -->
          <xsl:choose>
            <xsl:when test='$reserved_traffic &gt;= $traffic_limit'>

              <xsl:call-template name='limit'>
                <xsl:with-param name='limit' select='"max_traffic"'/>
                <xsl:with-param name='value' select='format-number($reserved_traffic + 1024*1024*1024, "#")'/>
              </xsl:call-template>

            </xsl:when>
            <xsl:otherwise>

              <xsl:call-template name='limit'>
                <xsl:with-param name='limit' select='"max_traffic"'/>
                <xsl:with-param name='value' select='$traffic_limit'/>
              </xsl:call-template>

            </xsl:otherwise>
          </xsl:choose>

        </xsl:if>

        <xsl:variable name='space_limit' select='format-number(limit[@name="disk_space"], "#")'/>

        <xsl:if test='($space_limit != "") and ($space_limit != -1) and ($space_limit &lt; $max_limit_value)'>

          <!-- count summary reserved limits -->
          <xsl:variable name='reserved_space' select='sum(account/limit[(@name="disk_space") and (. != -1) and (. != "") and (. &lt; $max_limit_value)])'/>

          <!-- if there is not enough resources to deploy accounts, which
               have their space reserved we should increase reseller's limit to at
               least the required value + 10Mb just in case... -->
          <xsl:choose>
            <xsl:when test='$reserved_space &gt;= $space_limit'>

              <xsl:call-template name='limit'>
                <xsl:with-param name='limit' select='"disk_space"'/>
                <xsl:with-param name='value' select='format-number($reserved_space + 1024*1024*10, "#")'/>
              </xsl:call-template>

            </xsl:when>
            <xsl:otherwise>

              <xsl:call-template name='limit'>
                <xsl:with-param name='limit' select='"disk_space"'/>
                <xsl:with-param name='value' select='$space_limit'/>
              </xsl:call-template>

            </xsl:otherwise>
          </xsl:choose>

        </xsl:if>

      </xsl:template>

      <!-- Some neccessary permissions, which exist on cPanel -->
      <xsl:template name="account-permissions">
		<!--Moved to template account-feature-permissions -->
        <!--permission name="manage_subdomains" allowed="true"/-->
        <!--permission name="manage_dns" allowed="true"/-->
        <!--permission name="manage_log" allowed="true"/-->
        <!--permission name="manage_crontab" allowed="true"/-->
        <!--permission name="manage_anonftp" allowed="true"/-->
        <!--permission name="manage_maillists" allowed="true"/-->
      </xsl:template>

      <!-- Handle reseller permissions -->
      <xsl:template name="reseller-permissions">
        <permission name="create_domains" allowed="true"/>
        <xsl:if test="./permission[@name='manage_dns']">
			<permission name="manage_dns" allowed="true"/>
        </xsl:if>
		<xsl:call-template name="account-permissions"/>
		<xsl:call-template name="account-feature-permissions"/>
      </xsl:template>
      
      <!-- Handle reseller/account features as plesk permissions-->
      <xsl:template name="account-feature-permissions">      
         <xsl:apply-templates select="./feature[@name!='password']"/>
      </xsl:template>
      
      <xsl:template name="feature-to-permissions" match="feature[@name]">
		  <xsl:choose>
					<xsl:when test="@name='addondomains' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_phosting</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>					
					</xsl:when>
					<xsl:when test="@name='backup' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">allow_local_backups</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>					
						<xsl:element name="permission">	
							<xsl:attribute name="name">allow_ftp_backups</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>					
					</xsl:when>
					<xsl:when test="@name='clamavconnector_scan' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_drweb</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>					
					</xsl:when>
					<xsl:when test="@name='cron' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_crontab</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>										
					</xsl:when>					
					<xsl:when test="@name='ftpsetup' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_anonftp</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>
					</xsl:when>					
					<xsl:when test="@name='lists' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_maillists</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>																
					</xsl:when>					
					<xsl:when test="@name='parkeddomains' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_domain_aliases</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>																
					</xsl:when>					
					<xsl:when test="@name='rawlog' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_log</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>																						
					</xsl:when>					
					<xsl:when test="@name='spamassassin' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_spamfilter</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>																						
					</xsl:when>					
					<xsl:when test="@name='statmanager' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_webstat</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>																												
					</xsl:when>					
					<xsl:when test="@name='subdomains' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_subdomains</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>																																		
					</xsl:when>										
					<xsl:when test="@name='password' ">
						<xsl:element name="permission">	
							<xsl:attribute name="name">manage_ftp_password</xsl:attribute>
							<xsl:attribute name="allowed"><xsl:value-of select="./@allowed"/></xsl:attribute>
						</xsl:element>																																		
					</xsl:when>															
				</xsl:choose>      				
      </xsl:template>

	<!--
	Convert cPanel domain into the Plesk one
	Params:
			quota - account's disk quota [optional]
			du_passwd - domain user password [optional]
	-->
	<xsl:template name="domain">
		<xsl:param name="quota"/>
		<xsl:param name="du_passwd"/>
		<xsl:param name="du_email"/>
		<xsl:param name="du_contact"/>

        <xsl:variable name="domainname" select="@name"/>

		<!-- Restore plain structure. Exceptions are handled later -->
		<domain name='{$domainname}'>
			<xsl:apply-templates select='../ip'/>

			<status>
				<enabled/>
			</status>

			<!-- disk space and traffic limits require special processing -->
			<xsl:call-template name='domain-limit-shared'>
				<xsl:with-param name='name' select='"max_traffic"'/>
			</xsl:call-template>
			<xsl:call-template name='domain-limit-shared'>
				<xsl:with-param name='name' select='"disk_space"'/>
			</xsl:call-template>

			<xsl:copy-of select='../limit[@name="max_subdom" and . != ""]'/>
			<xsl:copy-of select='../limit[@name="max_box" and . != ""]'/>
			<xsl:call-template name='limit'>
				<xsl:with-param name='limit' select='"max_wu"'/>
				<xsl:with-param name='value' select='../limit[@name="max_ftpusers"]'/>
			</xsl:call-template>

            <mailsystem>
              <status>
                <enabled/>
              </status>  
				<xsl:apply-templates select="mail"/>
			</mailsystem>

            <!-- parked domains -->
            <xsl:for-each select="addondomain">
                <xsl:call-template name="parked-domain">
                    <xsl:with-param name="domain" select="$domainname"/>
                </xsl:call-template>
            </xsl:for-each>
              
			<xsl:copy-of select="database"/>
			<xsl:apply-templates select="maillists"/>

			<xsl:call-template name='domainuser'>
				<xsl:with-param name='password' select='$du_passwd'/>
				<xsl:with-param name='cname' select='$du_contact'/>
				<xsl:with-param name='email' select='$du_email'/>
				<xsl:with-param name='level'>1</xsl:with-param>
			</xsl:call-template>

			<xsl:call-template name="hosting">
				<xsl:with-param name="quota" select="$quota"/>
			</xsl:call-template>
		</domain>

		<!--
		There is a very special issue about addon domains: they should
		be transformed into separate domains with standard forwarding
		pointing to Plesk's domains. Also there is an exception for those
		subdomains, which have only 1 addon domain - subdomain migrates
		into this one addon domain, thus making it domain with physical
		hosting.
        This is performed for addon domains.
        Parked domains are morphed into domain aliases
		-->

		<!-- usual addon domains -->
		<xsl:for-each select="subdomain[1!=count(addondomain) or not(ftpuser)]/addondomain">
			<xsl:call-template name="addondomain">
                    <xsl:with-param name="domain" select="concat(@name, '.', $domainname)"/>
				<xsl:with-param name='du_passwd' select='$du_passwd'/>
				<xsl:with-param name='du_email' select='$du_email'/>
				<xsl:with-param name='du_contact' select='$du_contact'/>
			</xsl:call-template>
		</xsl:for-each>

		<!-- addon-to-hosting domains -->
		<xsl:for-each select="subdomain[1 = count(addondomain) and ftpuser]">
			<xsl:call-template name='addondomain-to-hosting'>
				<xsl:with-param name='du_passwd' select='$du_passwd'/>
				<xsl:with-param name='du_cname' select='$du_contact'/>
				<xsl:with-param name='du_email' select='$du_email'/>
			</xsl:call-template>
		</xsl:for-each>

	</xsl:template>

	<!--
	cPanel domains always have hosting. Transform it into
	the Plesk's model. Generate the hosting element with all
	features enabled.
	Params: quota - if account has disk space limit [optional]
	-->
	<xsl:template name='hosting'>
		<xsl:param name='quota'/>

		<phosting https='true' fp='true' fpssl='true' fpauth='true' webstat='true' errdocs='true' shared-content='true'>
			<xsl:copy-of select="@cid_docroot[.!='']"/>
			<xsl:copy-of select="@cid_cgi[.!='']"/>

			<xsl:call-template name="sysuser">
				<xsl:with-param name="name"     select="ftpuser[not(@directory)]/@name"/>
				<xsl:with-param name="password" select="ftpuser[not(@directory)]/@password"/>
				<xsl:with-param name="quota"    select="$quota"/>
			</xsl:call-template>

			<xsl:call-template name="scripting-all"/>

			<fpuser name='{ftpuser[not(@directory)]/@name}'>
				<xsl:call-template name='password'>
					<xsl:with-param name='password' select='ftpuser[not(@directory)]/@password'/>
				</xsl:call-template>
			</fpuser>

			<xsl:copy-of select='anonftp'/>

			<!-- Protected directories -->
			<xsl:for-each select="pdir">
				<pdir>
					<xsl:copy-of select="@name[.!='']"/>
					<xsl:copy-of select="@title[.!='']"/>
					<xsl:for-each select="pduser">
						<pduser name="{@name}">
							<xsl:if test="@password">
								<password type="encrypted">
									<xsl:if test="@encoding">
										<xsl:attribute name="encoding">
											<xsl:value-of select="@encoding"/>
										</xsl:attribute>
									</xsl:if>
									<xsl:value-of select="@password"/>
								</password>
							</xsl:if>
						</pduser>
					</xsl:for-each>
				</pdir>
			</xsl:for-each>

			<!-- web users -->
			<xsl:apply-templates select="ftpuser[@directory]"/>
			<!-- consider subdomains, which shouldn't be converted to usual domains -->
			<xsl:apply-templates select="subdomain[1 != count(addondomain) or not(ftpuser)]"/>

		</phosting>
	</xsl:template>

	<!--
	Register subdomain which has no addon domain pointing on it
	(see additional details in comment to addondomain-to-hosting)
	with optional ftp user and all scripting enabled. No mail
	system however.
	-->
	<xsl:template name="subdomain" match="subdomain[1 != count(addondomain) or not(ftpuser)]">
		<subdomain>
			<xsl:copy-of select="@name[.!='']"/>
			<xsl:copy-of select="@cid_docroot[.!='']"/>
			<xsl:copy-of select="@cid_cgi[.!='']"/>

			<!-- may contain separate ftp user -->
			<xsl:if test="ftpuser">
				<xsl:call-template name="sysuser">
					<xsl:with-param name="name" select="ftpuser/@name"/>
					<xsl:with-param name="password" select="ftpuser/@password"/>
					<xsl:with-param name="quota" select="ftpuser/@quota"/>
				</xsl:call-template>
			</xsl:if>

			<xsl:call-template name="scripting-all"/>

		</subdomain>
	</xsl:template>

	<!--
	If subdomain has 1 addon domain assume that it's better to map this
	subdomain+addondomain pair into just domain with physical hosting
	and mail. Thus there would be no subdomain in the original domain.
	Transformation of subdomain into full blown domain with hosting is
	possible only when subdomain has separate ftp account - othervise we
	are unable to create physical hosting.
	Unfortunately we can't estimate subdomain's resource usage, so addon
	domain will have zero limits and therefore will be shown as 'problem'
	domain.

	Params:
			du_passwd - password for domain-level user [optional]
			du_cname - domain-level user contact name [optional]
			du_email - domain-level user email [optional]

	Use case:
		If the domain zzz.co.uk is setup as an addon pointing to the
		subdomain zzz.domain.com it should be transferred with the
		mail, and the contents of public_html/zzz should be transffered
		to the zzz.co.uk domain name on the Plesk server. This way, the
		subdomain is removed and it is treated as being a full (physical
		hosting) domain. In Plesk, both domains would have physical
		hosting and need no subdomains.
	-->
	<xsl:template name="addondomain-to-hosting" match="subdomain[1 = count(addondomain) and ftpuser]">
		<xsl:param name='du_passwd'/>
		<xsl:param name='du_cname'/>
		<xsl:param name='du_email'/>

		<domain name='{addondomain}'>
			<xsl:apply-templates select='../../ip'/>

			<status>
				<enabled/>
			</status>
			<!-- Limits. Count the required mailboxes limit. -->
			<xsl:call-template name='domain-limits-zero'>
				<xsl:with-param name='skip-mail' select='"yes"'/>
			</xsl:call-template>
			<xsl:call-template name='limit'>
				<xsl:with-param name='limit' select='"max_box"'/>
				<xsl:with-param name='value' select='count(mail/mailname)'/>
			</xsl:call-template>

            <mailsystem>
              <status>
                <enabled/>
              </status>  
				<xsl:apply-templates select='mail'/>
			</mailsystem>

			<xsl:call-template name='domainuser'>
				<xsl:with-param name='password' select='$du_passwd'/>
				<xsl:with-param name='cname' select='$du_cname'/>
				<xsl:with-param name='email' select='$du_email'/>
				<xsl:with-param name='level'>2</xsl:with-param>
			</xsl:call-template>

			<xsl:call-template name='hosting'/>
		</domain>
	</xsl:template>

    <!--
        Parked domains are "domain aliases" in Plesk
    -->
    <xsl:template name="parked-domain">
        <xsl:param name="domain"/>
        <xsl:variable name="domainalias" select='.'/>
        <domain-alias name="{$domainalias}" mail="true" web="true">
            <status>
                <enabled/>
            </status>
        </domain-alias>
    </xsl:template>

      <!--
           Convert addondomain into the Plesk's domain with the standard
           forwarding hosting.

           Parameters:
           domain - FQDN of the subdomain for the addon domain
           du_passwd  - domain user's password (if any)
           du_email   - domain user's email (if any)
           du_contact - domain user's contact name (if any)
           -->
      <xsl:template name="addondomain">

        <xsl:param name="domain"/>
        <xsl:param name="du_passwd"/>
        <xsl:param name="du_email"/>
        <xsl:param name="du_contact"/>

        <xsl:variable name="addondomain" select='.'/>

        <domain name="{$addondomain}">
		  <xsl:apply-templates select='../../../ip'/>

		  <status>
			<enabled/>
		  </status>
          <xsl:call-template name='domain-limits-zero'/>

		  <xsl:call-template name='domainuser'>
			<xsl:with-param name='password' select='$du_passwd'/>
			<xsl:with-param name='cname' select='$du_contact'/>
			<xsl:with-param name='email' select='$du_email'/>
			<xsl:with-param name='level'>3</xsl:with-param>
		  </xsl:call-template>

          <shosting>
            <xsl:value-of select="$domain"/>
          </shosting>
        </domain>

      </xsl:template>

      <!-- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
           Converts cPanel's ftp user into Plesk's web user. Such transformation
           may be performed only for ftp users of main domain, which have
           their own directories. Subdomains and directories within subdomains
           can't be transformed in such a way.
           Directory name doesn't matter in this case, so silently ignored...
           If webuser has no password - generate a random one...
           - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -->
      <xsl:template name="webuser" match="ftpuser">
        <xsl:element name="webuser">
          <xsl:copy-of select="@cid[.!='']"/>
          <xsl:choose>
            <xsl:when test='string-length(@password)=0'>
              <xsl:call-template name="sysuser">
                <xsl:with-param name="name"     select="@name"         />
                <xsl:with-param name="password" select="generate-id(.)"/>
                <xsl:with-param name="quota" select="@quota"/>
              </xsl:call-template>
            </xsl:when>
            <xsl:otherwise>
              <xsl:call-template name="sysuser">
                <xsl:with-param name="name"     select="@name"    />
                <xsl:with-param name="password" select="@password"/>
                <xsl:with-param name="quota" select="@quota"/>
              </xsl:call-template>
            </xsl:otherwise>
          </xsl:choose>
          <xsl:call-template name="scripting-all"/>
        </xsl:element>
      </xsl:template>

	<!--
	Outputs domain user element.
	Params: password - encrypted password [optional]
			cname - contact name [optional]
			email - contact email [optional]
	-->
	<xsl:template name='domainuser'>
		<xsl:param name='password'/>
		<xsl:param name='cname'/>
		<xsl:param name='email'/>
		<xsl:param name='level'/>

		<xsl:if test='string-length($password) > 0'>
			<domainuser>
				<xsl:if test="string-length($cname) > 0">
					<xsl:attribute name="contact">
						<xsl:value-of select="$cname"/>
					</xsl:attribute>
				</xsl:if>
				<xsl:call-template name="password">
					<xsl:with-param name="password" select="$password"/>
				</xsl:call-template>
				<xsl:if test='string-length($email) > 0'>
					<pinfo name='email'>
						<xsl:value-of select="$email"/>
					</pinfo>
				</xsl:if>
				<xsl:choose>
					<xsl:when test="$level=1"><xsl:apply-templates select="../feature"/></xsl:when>
					<xsl:when test="$level=2"><xsl:apply-templates select="../../feature"/></xsl:when>
					<xsl:when test="$level=3"><xsl:apply-templates select="../../../feature"/></xsl:when>
				</xsl:choose>
			</domainuser>
		</xsl:if>
	</xsl:template>

	<!--
	Outputs Plesk's sysuser element.
	Params: name - login name
			password - encrypted password
			quota - hard disk quota for this system user in bytes [optional]
	-->
	<xsl:template name="sysuser">
		<xsl:param name="name"/>
		<xsl:param name="password"/>
		<xsl:param name="quota"/>

		<sysuser name="{$name}">
			<xsl:if test="string-length($quota) > 0">
				<xsl:attribute name="quota">
					<xsl:value-of select="$quota"/>
				</xsl:attribute>
			</xsl:if>

			<xsl:call-template name="password">
				<xsl:with-param name="password" select="$password"/>
			</xsl:call-template>
		</sysuser>
	</xsl:template>

	<!--
	Outputs Plesk's password element.
	Params: password - encrypted || plaintext password
			password_type - password type [optional, default = 'encrypted']
	-->
	<xsl:template name="password">
		<xsl:param name="password"/>
		<xsl:param name="password_type">encrypted</xsl:param>

		<password>
          <xsl:attribute name="type">
            <xsl:value-of select="$password_type"/>
          </xsl:attribute>
          <xsl:value-of select="$password"/>
		</password>
	</xsl:template>

	<!--
	Outputs Plesk's scripting element with all scripting
	options enabled (used for webusers, hostings and subdomains).
	-->
	<xsl:template name="scripting-all">
		<scripting ssi="true" php="true" cgi="true" perl="true" python="true"/>
	</xsl:template>

	<!--
	Outputs Plesk's limit (used for domains and clients accounts)
	Params: limit - limit name
			value - limit value
	-->
	<xsl:template name="limit">
		<xsl:param name="limit"/>
		<xsl:param name="value"/>

		<limit name="{$limit}">
			<xsl:choose>
				<xsl:when test="string($value)='' or string($value)='Infinity' or string($value)='NaN' or string($value)='-Infinity'">-1</xsl:when>
				<xsl:otherwise>
					<xsl:value-of select="format-number($value, '#')"/>
				</xsl:otherwise>
			</xsl:choose>
		</limit>
	</xsl:template>

	<!-- Mail subsystem transformation -->
	<xsl:template name="mail" match="mail">

		<xsl:for-each select="mailname">
			<xsl:variable name="name" select="@name"/>

			<mailuser name="{$name}" mailgroup-enabled="false">
				<xsl:if test="@quota">
					<xsl:attribute name="mailbox-quota">
						<xsl:value-of select="@quota"/>
					</xsl:attribute>
				</xsl:if>
				<xsl:call-template name='password'>
					<xsl:with-param name='password' select='@password'/>
				</xsl:call-template>

				<mailbox type="mbox">
                    <xsl:copy-of select="@cid_inbox"/>
                    <xsl:if test="@cid">
                        <xsl:copy-of select="@cid"/>
                    </xsl:if>
				</mailbox>

				<!-- There may be forwarders and autoresponders in cPanel,
				which may be mapped on the Plesk's business model:
				mailname may have 1 forward and several autoresponders -->
				<xsl:if test="../forward[@mailname=$name]">
					<redirect>
						<xsl:value-of select="../forward[@mailname=$name]/@redirect"/>
					</redirect>
				</xsl:if>

				<xsl:if test="../autoresponder[@mailname=$name]">
					<autoresponders enabled="true">
						<xsl:apply-templates select="../autoresponder[@mailname=$name]"/>
					</autoresponders>
				</xsl:if>
							
				<xsl:choose>
					<xsl:when test="spamassassin">
						<xsl:copy-of select="spamassassin"/>
					</xsl:when>
					<xsl:otherwise>
						<xsl:copy-of select="../../../spamassassin"/>
					</xsl:otherwise>
				</xsl:choose>								
			</mailuser>
		</xsl:for-each>

		<!-- Convert separate autoresponders to loginless mailnames -->
		<xsl:for-each select="autoresponder">
			<xsl:variable name="name" select="@mailname"/>
			<!-- look that this autoresponder wasn't processed with mailname -->
			<xsl:if test="not(../mailname[@name=$name])">
				<mailuser name="{$name}" mailgroup-enabled="false">
					<autoresponders enabled="true">
						<xsl:apply-templates select="../autoresponder[@mailname=$name]"/>
					</autoresponders>
				</mailuser>
			</xsl:if>
		</xsl:for-each>

		<!-- Convert separate forwarders to loginless mailnames -->
		<xsl:for-each select="forward">
			<xsl:if test="not(../mailname[@name=current()/@name])">
				<xsl:choose>
					<xsl:when test="count(../forward[@mailname=current()/@mailname])=1">						
						<mailuser name="{@mailname}" mailgroup-enabled="false">
							<redirect>
								<xsl:value-of select="@redirect"/>
							</redirect>
						</mailuser>					
					</xsl:when>
					<xsl:otherwise>
							<xsl:if test="../forward[@mailname=current()/@mailname][1][@redirect=current()/@redirect]" >
								<mailuser name="{@mailname}" mailgroup-enabled="true">															
									<xsl:for-each select="../forward[@mailname=current()/@mailname]">
										<mailgroup-member>
											<xsl:value-of select="@redirect"/>
										</mailgroup-member>
									</xsl:for-each>
								</mailuser>					
							</xsl:if>
					</xsl:otherwise>
				</xsl:choose>
			</xsl:if>
		</xsl:for-each>	
		
		<!-- nonexistent mail user handling -->
		<xsl:if test="default">
			<catch-all>
				<xsl:if test="default/@target = 'email'">
					<xsl:value-of select="default"/>
				</xsl:if>
				<xsl:if test="default/@target = 'ignore'">reject</xsl:if>
				<xsl:if test="default/@target = 'fail'">
					<xsl:value-of select="concat('bounce:',default)"/>
				</xsl:if>
			</catch-all>
		</xsl:if>

	</xsl:template>

	<!-- Autoresponders transformation -->
	<xsl:template name="autoresponder" match="autoresponder">
		<xsl:variable name="name" select="@mailname"/>
		<autoresponder name="{$name}" status="on">
			<xsl:if test="@from">
				<xsl:attribute name="replyto">
				  <xsl:value-of select="@from"/>
				</xsl:attribute>
			</xsl:if>
			<xsl:copy-of select="@subject"/>
			<xsl:if test="@type">
				<xsl:attribute name="content-type">
					<xsl:choose>
						<xsl:when test="@type = 'plain'">text/plain</xsl:when>
						<xsl:otherwise>text/html</xsl:otherwise>
					</xsl:choose>
				</xsl:attribute>
			</xsl:if>
			<text>
				<xsl:copy-of select="@charset"/>
				<xsl:value-of select="."/>
			</text>
		</autoresponder>
	</xsl:template>	

	<!-- MailList transformation -->
	<xsl:template name="maillists" match="maillists">
		<maillists>
			<status><enabled/></status>
			<xsl:apply-templates select="maillist"/>
		</maillists>
	</xsl:template>	

	<xsl:template name="maillist" match="maillist">
		<xsl:element name="maillist">
			<xsl:attribute name="name">
				<xsl:value-of select="./@name"/>
			</xsl:attribute>
			<xsl:if test="@cid">
				<xsl:attribute name="cid">
					<xsl:value-of select="./@cid"/>
				</xsl:attribute>		
			</xsl:if>				
			<status><enabled/></status>
			<xsl:copy-of select="owner"/>
			<xsl:copy-of select="password"/>
			<xsl:copy-of select="recipient"/>
		</xsl:element>
	</xsl:template>	

</xsl:stylesheet>
